require 'pdf/writer'
require 'pdf/ohash'

module PDF
    # This class will take the basic interaction facilities of the PDF::Writer
    # class and make more useful functions so that the user does not have to
    # know all the ins and outs of PDF presentation to produce something
    # pretty.
  class EZWriter < Writer
    VERSION = "0.1.1"

    PAGE_SIZES = {
      "4A0"   => [0, 0, 4767.87, 6740.79], "2A0"    => [0, 0, 3370.39, 4767.87],
      "A0"    => [0, 0, 2383.94, 3370.39], "A1"     => [0, 0, 1683.78, 2383.94],
      "A2"    => [0, 0, 1190.55, 1683.78], "A3"     => [0, 0,  841.89, 1190.55],
      "A4"    => [0, 0,  595.28,  841.89], "A5"     => [0, 0,  419.53,  595.28],
      "A6"    => [0, 0,  297.64,  419.53], "A7"     => [0, 0,  209.76,  297.64],
      "A8"    => [0, 0,  147.40,  209.76], "A9"     => [0, 0,  104.88,  147.40],
      "A10"   => [0, 0,   73.70,  104.88], "B0"     => [0, 0, 2834.65, 4008.19],
      "B1"    => [0, 0, 2004.09, 2834.65], "B2"     => [0, 0, 1417.32, 2004.09],
      "B3"    => [0, 0, 1000.63, 1417.32], "B4"     => [0, 0,  708.66, 1000.63],
      "B5"    => [0, 0,  498.90,  708.66], "B6"     => [0, 0,  354.33,  498.90],
      "B7"    => [0, 0,  249.45,  354.33], "B8"     => [0, 0,  175.75,  249.45],
      "B9"    => [0, 0,  124.72,  175.75], "B10"    => [0, 0,   87.87,  124.72],
      "C0"    => [0, 0, 2599.37, 3676.54], "C1"     => [0, 0, 1836.85, 2599.37],
      "C2"    => [0, 0, 1298.27, 1836.85], "C3"     => [0, 0,  918.43, 1298.27],
      "C4"    => [0, 0,  649.13,  918.43], "C5"     => [0, 0,  459.21,  649.13],
      "C6"    => [0, 0,  323.15,  459.21], "C7"     => [0, 0,  229.61,  323.15],
      "C8"    => [0, 0,  161.57,  229.61], "C9"     => [0, 0,  113.39,  161.57],
      "C10"   => [0, 0,   79.37,  113.39], "RA0"    => [0, 0, 2437.80, 3458.27],
      "RA1"   => [0, 0, 1729.13, 2437.80], "RA2"    => [0, 0, 1218.90, 1729.13],
      "RA3"   => [0, 0,  864.57, 1218.90], "RA4"    => [0, 0,  609.45,  864.57],
      "SRA0"  => [0, 0, 2551.18, 3628.35], "SRA1"   => [0, 0, 1814.17, 2551.18],
      "SRA2"  => [0, 0, 1275.59, 1814.17], "SRA3"   => [0, 0,  907.09, 1275.59],
      "SRA4"  => [0, 0,  637.80,  907.09], "LETTER" => [0, 0,  612.00,  792.00],
      "LEGAL" => [0, 0,  612.00, 1008.00], "FOLIO"  => [0, 0,  612.00,  936.00],
      "EXECUTIVE" => [0, 0,  521.86,  756.00]
    }

      # Assuming that people don't want to specify the paper size using the
      # absolute coordinates allow a couple of options:
      # +paper+::       specifies a page size name
      # +orientation+:: :portrait or :landscape
      #
      # To actually set the coordinates, then pass an array in as the first
      # parameter. the defaults are as shown. If the array is only two values
      # (instead of the 4 for coordinates), it represents the page width and
      # height in centimeters.
    def initialize(paper = "A4", orientation = :portrait)
      @ez = { :fontSize => 10 }
      @y = nil
      @ezPages = []

      if paper.kind_of?(Array)
        if paper.size == 4
          size = paper # Coordinate Array
        else
          size = [0, 0, cm2pts(paper[0]), cm2pts(paper[1])]
            # Paper size in centimeters has been passed
        end
      else
        size = PAGE_SIZES[paper.upcase].dup
        size[3], size[2] = size[2], size[3] if orientation == :landscape
      end

      super(size)
      @ez[:pageWidth] = size[2]
      @ez[:pageHeight] = size[3]

        # also set the margins to some reasonable defaults
      @ez[:topMargin]    = 30;
      @ez[:bottomMargin] = 30;
      @ez[:leftMargin]   = 30;
      @ez[:rightMargin]  = 30;

        # Set the current writing position to the top of the first page
      @y = @ez[:pageHeight] - @ez[:topMargin]
        # Get the ID of the page that was created during the instantiation
        # process.
      @ezPages[1] = @pages.first_page
    end

    def ez_pages
      @ezPages
    end

    attr_reader :ez
    attr_accessor :y

    def cm2pts(x)
      (x / 2.54) * 72
    end

    def ez_set_cm_margins(top, bottom, left, right)
      ez_set_margins(cm2pts(top), cm2pts(bottom), cm2pts(left), cm2pts(right))
    end

    def ez_columns_start(options = nil)
        # Start from the current y-position; make the set number of columns.
      unless @ez[:columns].nil?
        return if @ez[:columns] == 1
      end

      options ||= {}
      { :gap => 10, :num => 2 }.each { |k, v| options[k] = v }

      @ez[:columns] ||= { :on => true, :colNum => 1 }
      @ez[:columns][:on] = true
        # store the current margins
      @ez[:columns][:margins] = [@ez[:leftMargin], @ez[:rightMargin], @ez[:topMargin], @ez[:bottomMargin]]
        # and store the settings for the columns
      @ez[:columns][:options] = options;
        # Reset the margins to suit the new columns. Safe enough to assume the
        # first column here, but start from the current y-position.
      @ez[:topMargin] = @ez[:pageHeight] - @y
      width = (@ez[:pageWidth] - @ez[:leftMargin] - @ez[:rightMargin] - (options[:num] - 1) * options[:gap]) / options[:num]
      @ez[:columns][:width] = width
      @ez[:rightMargin] = @ez[:pageWidth] - @ez[:leftMargin] - width
    end

    def ez_columns_stop
      if not @ez[:columns].nil? and @ez[:columns][:on]
        @ez[:columns][:on] = false
        @ez[:leftMargin], @ez[:rightMargin], @ez[:topMargin], @ez[:bottomMargin] = @ez[:columns][:margins]
      end
    end

    def ez_insert_mode(on = true, page_num = nil, pos = :before)
      if on
        unless @ezPages[page_num].nil?
          @ez[:insertMode] = true
          @ez[:insertOptions] = { :id => @ezPages[page_num], :pos => pos }
        end
      else
        @ez[:insertMode] = false
      end
    end

    def ez_new_page
      page_required = true
      if not @ez[:columns].nil? and @ez[:columns][:on]
          # Check if this is just going to a new column. Increment the column
          # number.
        @ez[:columns][:colNum] += 1
        if @ez[:columns][:colNum] <= @ez[:columns][:options][:num]
          page_required = false
        else
          @ez[:columns][:colNum] = 1
          @ez[:topMargin] = @ez[:columns][:margins][2]
        end

        width = @ez[:columns][:width]
        @ez[:leftMargin] = @ez[:columns][:margins][0] + (@ez[:columns][:colNum] - 1) * (@ez[:columns][:options][:gap] + width)
        @ez[:rightMargin] = @ez[:pageWidth] - @ez[:leftMargin] - width
      end

      if page_required
          # make a new page, setting the writing point back to the top.
        @y = @ez[:pageHeight] - @ez[:topMargin]
          # make the new page with a call to the basic class
        if @ez[:insertMode]
          id = new_page(true, @ez[:insertOptions][:id], @ez[:insertOptions][:pos])
          @ezPages << id
            # Manipulate the insert options so that inserted pages follow each
            # other
          @ez[:insertOptions][:id] = id
          @ez[:insertOptions][:pos] = :after
        else
          @ezPages << new_page
        end
      else
        @y = @ez[:pageHeight] - @ez[:topMargin]
      end
    end

    def ez_set_margins(top, bottom, left, right)
        # Set the margins to new values
      @ez[:topMargin]    = top
      @ez[:bottomMargin] = bottom
      @ez[:leftMargin]   = left
      @ez[:rightMargin]  = right
        # Check to see if this means that the current writing position is
        # outside the writable area
      if @y > (@ez[:pageHeight] - top)
          # Move y down
        @y = @ez[:pageHeight] - top
      end

      ez_new_page if @y < bottom # Make a new page
    end

    def ez_get_current_page_number
      @ezPages.size
    end

      # Put page numbers on the pages from here. Place them on the +pos+ side
      # of the coordinates (x, y).
      #
      # +pos+ can be :left or :right.
      #
      # Use the given +pattern+ for display, where <PAGENUM> and
      # <TOTALPAGENUM> are replaced as required. (Default is "<PAGENUM> of
      # <TOTALPAGENUM>".)
      #
      # If +num+ is set, this is the first page number; the number of total
      # pages will be adjusted to account for this.
      #
      # Adjust this function so that each time you :start page numbers then
      # you effectively start a different batch. Return the number of the
      # batch, so that they can be stopped in a different order if required.
    def ez_start_page_numbers(x, y, size, pos = nil, pattern = nil, num = nil)
      pos = :left if pos.nil? or pos.empty?
      pattern = "<PAGENUM> of <TOTALPAGENUM>" if pattern.nil? or pattern.empty?
      (@ez[:pageNumbering] ||= []) << (o = {})
      o[@ezPages.size] = { :x => x, :y => y, :pos => pos, :pattern => pattern, :num => num, :size => size }
    end

      # Given a particular generic page number (ie, document numbered
      # sequentially from beginning), return the page number under a
      # particular page numbering scheme (i)
    def ez_what_page_number(page_num, i = 0)
      num = 0
      start = start_num = 1
      @ez[:pageNumbering][i].each do |k, v|
        if k <= page_num
          if v.kind_of?(Hash)
            unless v[:num].nil?
              start = v[:num]
              start_num = k
              num = page_num - start_num + start
            end
          else
            num = 0
          end
        end
      end
      num
    end

      # Stop page numbering. If +num_block+ is set, stop that particular page
      # numbering sequence. If +stop_total+ is +true+, the totalingg of pages
      # for this numbering block will stop, too. If +stop_now+ is +true+, then
      # this page and all following pages will not be numbered; otherwise only
      # following pages will not be numbered.
    def ez_stop_page_numbers(stop_total = false, stop_now = true, num_block = 0)
      @ez[:pageNumbering] = {} if @ez[:pageNumbering].nil?

      if not stop_now and @ez[:pageNumbering][num_block][@ezPages.size].kind_of?(Hash)
          # This has only just been started; this will overwrite the start and
          # nothing will appear. Add a special command to the to the start
          # block, telling it to stop as well.
        if (stop_total)
          @ez[:pageNumbering][num_block][@ezPages.size]["stoptn"] = true
        else
          @ez[:pageNumbering][num_block][@ezPages.size]["stopn"] = true
        end
      else
        if (stop_total)
          @ez[:pageNumbering][num_block][@ezPages.size] = "stopt"
        else
          @ez[:pageNumbering][num_block][@ezPages.size] = "stop"
        end
        @ez[:pageNumbering][num_block][@ezPages.size] << "n" unless stop_now
      end
    end

    def ez_prvt_page_number_search(label, tmp)
      tmp.each do |i, v|
        if v.kind_of?(Hash)
          return i unless v[label].nil?
        else
          return i if v == label
        end
      end
      0
    end

    def ez_prvt_add_page_numbers
        # This will go through the pageNumbering array and add the page numbers
        # are required.
      unless @ez[:pageNumbering].nil?
        total_pages1 = @ezPages.size
        tmp1 = @ez[:pageNumbering]
        status = 0
        info = {}
        tmp1.each do |tmp|
            # Do each of the page numbering systems. First, find the total
            # pages for this one.
          k = ez_prvt_page_number_search("stopt", tmp)
          if not k.nil? and k > 0
            total_pages = k - 1
          else
            l = ez_prvt_page_number_search("stoptn", tmp)
            if not l.nil? and l > 0
              total_pages = l
            else
              total_pages = total_pages1
            end
          end
          @ezPages.each_with_index do |id, page_num|
            next if page_num == 0
            if tmp[page_num].kind_of?(Hash) # This must be the starting page #s
              status = 1
              info = tmp[page_num]
              info[:dnum] = info[:num] - page_num
                # Also check for the special case of the numbering stopping
                # and starting on the same page.
              status = 2 if not info["stopn"].nil? or not info["stoptn"].nil?
            elsif tmp[page_num] == "stop" or tmp[page_num] == "stopt"
              status = 0 # we are stopping page numbers
            elsif status == 1 and (tmp[page_num] == "stoptn" or tmp[page_num] == "stopn")
              status = 2
            end

            if status != 0
                # Add the page numbering to this page
              if info[:dnum].nil?
                num = page_num
              else
                num = page_num + info[:dnum]
              end

              total = total_pages + num - page_num
              pat = info[:pattern].gsub(/<PAGENUM>/, num.to_s).gsub(/<TOTALPAGENUM>/, total.to_s)
              reopen_object(id)
              case info[:pos]
              when :right
                add_text(info[:x], info[:y], info[:size], pat)
              else
                w = get_text_width(info[:size], pat)
                add_text(info[:x] - w, info[:y], info[:size], pat)
              end
              close_object
              status = 0 if status == 2
            end
          end
        end
      end
    end

    alias ez_prvt_clean_up ez_prvt_add_page_numbers

    def ez_output(options = 0)
      ez_prvt_clean_up
      to_s(options)
    end

      # Used to change the vertical position of the writing point.
    def ez_set_y(y)
      @y = y
      ez_new_page if @y < @ez[:bottomMargin]
    end

      # Used to change the vertical position of the writing point. Changes up
      # by a positive increment, so enter a negative number to go down the
      # page. If mod is set to :makeSpace and a new page is forced, then the
      # pointer will be moved down on the new page, this will allow space to
      # be reserved for graphics, etc.
    def ez_set_dy(dy, mod = nil)
      @y += dy
      if @y < @ez[:bottomMargin]
        ez_new_page
        @y += dy if mod == :makeSpace
      end
    end

    def ez_prvt_table_draw_lines(pos, gap, x0, x1, y0, y1, y2, col, inner, outer, opt = 1)
      x0 = 1000
      x1 = 0

      set_stroke_color(col)

      cnt = 0
      n = pos.size

      pos.each do |name, x|
        cnt += 1

        if (cnt == 1 or cnt == n)
          set_line_style(outer)
        else
          set_line_style(inner)
        end

        line(x - gap / 2, y0, x - gap / 2, y2)
        x1 = x if x > x1
        x0 = x if x < x0
      end

      set_line_style(outer)

      line(x0 - gap / 2 - outer / 2, y0, x1 - gap / 2 + outer / 2, y0)
        # Only do the second line if it is different than the first AND each
        # row does not have a line on it.
      line(x0 - gap / 2, y1, x1 - gap / 2, y1) if (y0 != y1 and opt < 2)
      line(x0 - gap / 2 - outer / 2, y2, x1 - gap / 2 + outer / 2, y2)
    end

        # Uses ezText to add the text, and returns the height taken by the
        # largest heading. This page will move the headings to a new page if
        # they will not fit completely on this one transaction support will be
        # used to implement this.
    def ez_prvt_table_column_headings(cols, pos, max_width, height, descender, gap, size, y, options_all = nil)
      options = {} if options.nil?
      options = options_all[:cols] unless options_all[:cols].nil?

      mx = second_go = 0
      start_page = @ezPages.size

        # y is the position at which the top of the table should start, so the
        # base of the first text, is y-height-gap-descender, but ezText starts
        # by dropping height.

        # The return from this function is the total cell height, including
        # gaps, and y is adjusted to be the postion of the bottom line.
      start_transaction(:column_headings)
      ok = false
      y -= gap
      loop do
        break if ok
        cols.each do |name, heading|
          ez_set_y(y)
          if not options[name].nil? and not options[name][:justification].nil?
            justification = options[name][:justification]
          else
            justification = :left
          end
          ez_text(heading, size, :aleft => pos[name],
                  :aright => (max_width[name] + pos[name]),
                  :justification => justification)
          dy = y - @y
          mx = dy if dy > mx
        end

        y -= (mx + gap) - descender # y = y - mx - gap + descender

          # If this has been moved to a new page, then abort the transaction;
          # move to a new page, and put it there. Do not check on the second
          # time around to avoid an infinite loop.
        if (@ezPages.size != start_page and not second_go)
          rewind_transaction(:column_headings)
          ez_new_page
          save_state
          y = @y - gap - descender
          ok = false
          second_go = true
          mx = 0
        else
          commit_transaction(:column_headings)
          ok = true
        end
      end

      return [mx + gap * 2 - descender, y]
    rescue
      abort_transaction(:column_headings)
    end

      # Will calculate the maximum width, taking into account that the text
      # may be broken by line breaks.
    def get_text_block_width(size, text)
      mx = 0
      text.each do |line|
        w = get_text_width(size, line)
        mx = w if w > mx
      end
      mx
    end

    TABLE_DEFAULTS = {
      :shaded              => 1,
      :showLines           => 1,
      :shadeCol            => PDF::Color.new(0.8, 0.8, 0.8),
      :shadeCol2           => PDF::Color.new(0.7, 0.7, 0.7),
      :fontSize            => 10,
      :titleFontSize       => 12,
      :titleGap            => 5,
      :lineCol             => PDF::Color.new(0, 0, 0),
      :gap                 => 5,
      :xPos                => :centre,
      :xOrientation        => :centre,
      :showHeadings        => true,
      :textCol             => PDF::Color.new(0, 0, 0),
      :width               => 0,
      :max_width            => 0,
      :cols                => PDF::OHash.new,
      :minRowSpace         => -100,
      :rowGap              => 2,
      :colGap              => 5,
      :innerLineThickness  => 1,
      :outerLineThickness  => 1,
      :splitRows           => false,
      :protectRows         => 1
    }

    Struct.new("Point", :x, :y)
    Struct.new("Box", :ul, :br)

    WIDTH_FACTOR = 1.01

      # Add a table of information to the pdf document.
      # [+data+]    A two dimensional Hash of the data to be presented.
      # [+cols+]    (optional) is a Hash; the keys are the names of the
      #             columns from +data+ to be presented (and in that order),
      #             the values are the titles to be given to the columns.
      # [+title+]   (optional) The title to be put on the top of the table.
      # [+options+] A Hash containing:
      #
      # :showLines::          0 (no lines), 1 (outside and top lines,
      #                       *default*), 2 (lines on each row)
      # :showHeadings::       false (no headings) or true (show headings)
      # :shaded::             0 (no shading), 1 (alternate lines shaded,
      #                       *default*), 2 (both shaded, second uses
      #                       shadeCol2), 3 ???
      # :shadeCol::           PDF::Color, defining the colour of the shading;
      #                       *default* (0.8, 0.8, 0.8)
      # :shadeCol2::          PDF::Color, defining the colour of the shading;
      #                       *default* (0.7, 0.7, 0.7)
      # :fontSize::           points, *default* 10
      # :textCol::            PDF::Color, defining text colour. *default*
      #                       black (0, 0, 0).
      # :titleFontSize::      points, *default* 12
      # :rowGap::             the space added at the top and bottom of each
      #                       row, between the text and the lines; *default* 2
      # :colGap::             the space on the left and right sides of each
      #                       cell; *default* 5
      # :lineCol::            PDF::Color, defining line colours. *default*
      #                       black (0, 0, 0).
      # :xPos::               :left, :right, :center, :centre, or a reference
      #                       coordinate in the x-direction.
      # :xOrientation::       :left, :right, :center, :centre, position of the
      #                       table wrt :xPos
      # :width::              <number> specifies the width of the table, if it
      #                       turns out to not be this wide, then it will
      #                       stretch the table to fit, if it is wider then
      #                       each cell will be made proportionalty smaller,
      #                       and the content may have to wrap.
      # :max_width::           <number> similar to :width, but will only make
      #                       table smaller than it wants to be
      # :options::            Hash of :colName => { :justification => :left,
      #                       :width => 100, :link => linkDataName}, ...;
      #                       allows the setting of other parameters for
      #                       individual columns.
      # :minRowSpace::        the minimum space between the bottom of each
      #                       row and the bottom margin, in which a new row
      #                       will be started if it is less, then a new page
      #                       would be started, *default* -100
      # :innerLineThickness:: *default* 1
      # :outerLineThickness:: *default* 1
      # :splitRows::          whether or not to allow the rows to be split
      #                       across page boundaries (+false+ not allowed
      #                       *default*, +true+ allowed)
      # :protectRows::        number, the number of rows to hold with the
      #                       heading on page. If there less than this number
      #                       of rows on the page, then move the whole lot
      #                       onto the next page. *default* 1
      #
      # Note that the user will have had to make a font selection already or
      # this will not produce a valid PDF file.
    def ez_table(data, cols = nil, title = nil, options = nil)
      raise :data_must_be_array unless data.kind_of?(Array)

      low_y   = descender = y0 = y1 = nil
        # This defines the "box" of the table.
      box     = Struct::Box.new
      box.ul  = Struct::Point.new
      box.br  = Struct::Point.new

        # Take the display columns as an array from the first row of the data
        # set. This is required because Ruby hashes are not ordered.
      col_order = data.shift
      raise :columns_unspecified unless col_order.kind_of?(Array)

      if cols.nil?
        cols = PDF::OHash.new
        col_order.each { |k| cols[k] = k }
      else
        o, cols = cols, PDF::OHash.new

        col_order.each { |k| cols[k] = o[k] }
        o = nil
      end

        # Set the options.
      options ||= {}
      TABLE_DEFAULTS.each { |k, v| options[k] = v if options[k].nil? }
      options[:gap] = 2 * options[:colGap]

        # Determine the middle of the page.
      middle = (@ez[:pageWidth] - @ez[:rightMargin]) / 2 + (@ez[:leftMargin] / 2)

      max_width = __find_table_max_width(data, cols, options)
      pos, t, x, adjustment_width, set_width = __find_table_positions(max_width, options)

        # if max_width is specified, and the table is too wide, and the width
        # has not been set, then set the width.
      options[:width] = options[:max_width] if options[:width] == 0 and options[:max_width] and ((t - x) > options[:max_width])

      if options[:width] and (adjustment_width > 0) and (set_width < options[:width])
          # first find the current widths of the columns involved in this
          # mystery
        cols0 = PDF::OHash.new
        cols1 = PDF::OHash.new

        xq = presentWidth = 0
        last = nil

        pos.each do |name, p|
          if options[:cols][last].nil? or
             options[:cols][last][:width].nil? or
             options[:cols][last][:width] <= 0
            unless last.nil? or last.empty?
              cols0[last] = p - xq - options[:gap]
              presentWidth += (p - xq - options[:gap])
            end
          else
            cols1[last] = p - xq
          end
          last = name
          xq = p
        end

          # cols0 contains the widths of all the columns which are not set
        needed_width = options[:width] - set_width

          # if needed width is negative then add it equally to each column,
          # else get more tricky
        if presentWidth < needed_width
          diff = (needed_width - presentWidth) / cols0.size
          cols0.each_key { |name| cols0[name] += diff }
        else
          cnt = 0
          loop do
            break if (presentWidth <= needed_width) or (cnt >= 100)
            cnt += 1 # insurance policy
              # Find the widest columns and the next to widest width
            aWidest = []
            nWidest = widest = 0
            cols0.each do |name, w|
              if w > widest
                aWidest << name
                nWidest = widest
                widest = w
              elsif w == widest
                aWidest << name
              end
            end

              # Then figure out what the width of the widest columns would
              # have to be to take up all the slack.
            newWidestWidth = widest - (presentWidth - needed_width) / aWidest.size
            if newWidestWidth > nWidest
              aWidest.each { |name| cols0[name] = newWidestWidth }
              presentWidth = needed_width
            else
                # There is no space, reduce the size of the widest ones down
                # to the next size down, and we will go round again
              aWidest.each { |name| cols0[name] = nWidest }
              presentWidth -= (widest - nWidest) * aWidest.size
            end
          end
        end

          # cols0 now contains the new widths of the constrained columns. now
          # need to update the pos and max_width arrays
        xq = 0
        pos.each do |name, p|
          pos[name] = xq

          if options[:cols][name].nil? or
             options[:cols][name][:width].nil? or
             options[:cols][name][:width] <= 0
            if not cols0[name].nil?
              xq += cols0[name] + options[:gap]
              max_width[name] = cols0[name]
            end
          else
            xq += cols1[name] unless cols1[name].nil?
          end
        end

        t = x + options[:width]
        pos[:__end__] = t
      end
      
        # now adjust the table to the correct location across the page
      case options[:xPos]
      when :left
        xref = @ez[:leftMargin]
      when :right
        xref = @ez[:pageWidth] - @ez[:rightMargin]
      when :center, :centre
        xref = middle
      else
        xref = options[:xPos]
      end

      case options[:xOrientation]
      when :left
        dx = xref - t
      when :right
        dx = xref
      when :center, :centre
        dx = xref - (t / 2)
      end

      pos.each { |k, v| pos[k] = v + dx }

      box.ul.x = x0 = x + dx
      box.br.x = x1 = t + dx

      baseLeftMargin = @ez[:leftMargin]
      basePos = pos.dup

        # ok, just about ready to make me a table
      set_color(options[:textCol])
      set_stroke_color(options[:shadeCol])

      middle = (x0 + x1) / 2

        # Start a transaction. This transaction will be used to regress the
        # table if there are not enough rows protected. 
      start_transaction(:table)
      moved_once = false if options[:protectRows].nonzero?

      abortTable = true
      loop do # while abortTable
        break unless abortTable
        abortTable = false

          # If the title is set, then set it
        unless title.nil? or title.empty?
          w = get_text_width(options[:titleFontSize], title)
          @y -= get_font_height(options[:titleFontSize])
          if @y < @ez[:bottomMargin]
            ez_new_page
            save_state
            @y -= get_font_height(options[:titleFontSize])
          end
          add_text(middle - w / 2, @y, options[:titleFontSize], title)
          @y -= options[:titleGap]
        end

        y = @y  # simplifies the code a bit
        low_y = y if low_y.nil? or y < low_y 

          # make the table
        height = get_font_height(options[:fontSize])
        descender = get_font_descender(options[:fontSize])

        y0 = y + descender
        dy = 0

        if options[:showHeadings]
            # This function will move the start of the table to a new page if
            # it does not fit on this one.
          headingHeight, y = ez_prvt_table_column_headings(cols, pos,
              max_width, height, descender, options[:rowGap],
              options[:fontSize], y, options)
          y0 = y + headingHeight
          y1 = y
        else
          y1 = y0
        end

        first_line = true

          # open an object here so that the text can be put in over the
          # shading
        tOID = __open_new_object if (options[:shaded] != 0)

        cnt = 0
        newPage = false
        data.each do |row|
          cnt += 1
            # Start a transaction that will be used for this row to prevent it
            # from being split.
          unless options[:splitRows]
            pageStart = @ezPages.size
            start_transaction(:row)
            row_orig = row
            y_orig = y
            y0_orig = y0
            y1_orig = y1
          end # unless options[:splitRows]

          ok = false
          second_turn = false
          loop do # while !abortTable and !ok
            break if abortTable or ok

            mx = 0
            newRow = true

            loop do # while !abortTable and (newPage or newRow)
              break if abortTable or not (newPage or newRow)

              y -= height
              low_y = y if low_y.nil? or y < low_y 

              if newPage or
                 (y < @ez[:bottomMargin]) or
                 (not options[:minRowSpace].nil? and
                  y < (@ez[:bottomMargin] + options[:minRowSpace]))
                  # check that enough rows are with the heading
                moved_once = abortTable = true if options[:protectRows] > 0 and not moved_once and cnt <= options[:protectRows]

                y2 = y - mx + (2 * height) + descender - (newRow ? 1 : 0) * height

                if options[:showLines] != 0
                  y0 = y1 if not options[:showHeadings]
                  ez_prvt_table_draw_lines(pos, options[:gap], x0, x1, y0, y1,
                      y2, options[:lineCol], options[:innerLineThickness],
                      options[:outerLineThickness], options[:showLines])
                end

                if options[:shaded] != 0
                  close_object
                  restore_state
                end

                ez_new_page
                save_state

                  # and the margins may have changed
                dm = @ez[:leftMargin] - baseLeftMargin
                basePos.each { |k, v| pos[k] = v + dm }

                box.ul.x = x0 = x0 + dm
                box.br.x = x1 = x1 + dm

                tOID = __open_new_object if options[:shaded] != 0

                set_color(options[:textCol], true)
                y = @ez[:pageHeight] - @ez[:topMargin]
                low_y = y
                y0 = y + descender
                mx = 0

                if options[:showHeadings]
                  ez_prvt_table_column_headings(cols, pos, max_width, height,
                      descender, options[:rowGap], options[:fontSize], y,
                      options)
                  y1 = y
                else
                  y1 = y0
                end

                first_line = true
                y -= height
                low_y = y if low_y.nil? or y < low_y 
              end
              
              newRow = false

                # Write the actual data. If these cells need to be split over
                # a page, then newPage will be set, and the remaining text
                # will be placed in leftOvers
              newPage = false
              leftOvers = PDF::OHash.new

              cols.each do |name, title|
                ez_set_y(y + height)
                colNewPage = false

                unless row[name].nil?
                  lines = row[name].to_s.split("\n")
                  if not options[:cols].nil? and
                     not options[:cols][name].nil? and
                     not options[:cols][name][:link].nil?
                    lines.map! { |k| k = "<c:alink:#{row[options[:cols][name][:link]]}>#{k}</c:alink>" }
                  end
                else
                  lines = []
                end

                @y -= options[:rowGap]

                lines.each do |line|
                  ez_process_text(line)
                  start = true

                  loop do
                    break if (line.nil? or line.empty?) and not start
                    start = false

                    @y = @y - height if not colNewPage

                      # a new page is required
                    newPage = colNewPage = true if @y < @ez[:bottomMargin]

                    if colNewPage
                      if leftOvers[name].nil?
                        leftOvers[name] = [line]
                      else
                        leftOvers[name] << "\n#{line}"
                      end
                      line = nil
                    else
                      if not options[:cols][name].nil? and
                         not options[:cols][name][:justification].nil?
                        just = options[:cols][name][:justification]
                      else
                        just = :left
                      end

                      line = add_text_wrap(pos[name], @y, max_width[name], options[:fontSize], line, just)
                    end
                  end
                end

                dy = y + height - @y + options[:rowGap]
                mx = dy - height * (newPage ? 1 : 0) if (dy - height * (newPage ? 1 : 0)) > mx
              end

                # Set row to leftOvers so that they will be processed onto the
                # new page
              row = leftOvers

                # Now add the shading underneath
              if options[:shaded] > 0
                close_object
                if (cnt % 2) == 0
                  set_color(options[:shadeCol], true)
                  filled_rectangle(x0 - options[:gap] / 2, y + descender + height - mx, x1 - x0, mx)
                elsif (cnt % 2) == 1 and options[:shaded] == 2
                  set_color(options[:shadeCol2], true)
                  filled_rectangle(x0 - options[:gap] / 2, y + descender + height - mx, x1 - x0, mx)
                end
                reopen_object(tOID)
              end

              if (options[:showLines] > 1)
                  # draw a line on the top of the block
                save_state
                set_stroke_color(options[:lineCol], true)
                if firstLine
                  set_line_style(options[:outerLineThickness])
                  firstLine = false
                else
                  set_line_style(options[:innerLineThickness])
                end
                line(x0 - options[:gap] / 2, y + descender + height, x1 - options[:gap] / 2, y + descender + height)
                restore_state
              end
            end

            y = y - mx + height
            low_y = y if low_y.nil? or y < low_y 

              # checking row split over pages
            unless options[:splitRows]
              if @ezPages.size != pageStart and not second_turn
                  # then we need to go back and try that again!
                newPage = second_turn = true
                rewind_transaction(:row)
                row = row_orig
                low_y = y = y_orig
#               low_y = y if low_y.nil? or y < low_y 
                y0 = y0_orig
                y1 = y1_orig
                ok = false
              else
                commit_transaction(:row)
                ok = true
              end
            else
              ok = true # don't go 'round the loop if splitting rows is allowed
            end
          end

          if abortTable
            # abort_transaction if not ok
              # only the outer transaction should be operational
            rewind_transaction(:table)
            ez_new_page
            save_state
            break
          end
        end
      end

      if low_y <= y
        y2 = low_y + descender
      else
        y2 = y + descender
      end

      if options[:showLines] != 0
        y0 = y1 if not options[:showHeadings]

        ez_prvt_table_draw_lines(pos, options[:gap], x0, x1, y0, y1, y2,
                                 options[:lineCol],
                                 options[:innerLineThickness],
                                 options[:outerLineThickness],
                                 options[:showLines])
      end

        # close the object for drawing the text on top
      if options[:shaded] != 0
        close_object
        restore_state
      end

      @y = low_y

        # Table has been put on the page, the rows guarded as required;
        # commit.
      commit_transaction(:table)

      [y]
    rescue
      abort_transaction(:table)
      raise
    end

      # Find the maximum widths of the text within each column. Default to
      # zero.
    def __find_table_max_width(data, cols, options)
      max_width = PDF::OHash.new(0)

        # Find the maximum cell widths based on the data and the headings.
        # Passing through the data multiple times is unavoidable as we must do
        # some analysis first.
      first = true
      data.each do |row|
        cols.each do |name, display|
          w = get_text_block_width(options[:fontSize], row[name].to_s) * WIDTH_FACTOR
          max_width[name] = w if w > max_width[name]
          if first
            w = get_text_block_width(options[:fontSize], display.to_s) * 1.01
            max_width[name] = w if w > max_width[name]
          end
        end
        first = false
      end
      max_width
    end

      # Calculate the start positions of each of the columns. This is based
      # on max_width, but may be modified with column options.
    def __find_table_positions(max_width, options)
      pos = PDF::OHash.new
      x = t = adjustment_width = set_width = 0

      max_width.each do |name, w|
        pos[name] = t
          # If the column width has been specified then set that here, also
          # total the width avaliable for adjustment.
        if not options[:cols][name].nil? and
           not options[:cols][name][:width].nil? and
           options[:cols][name][:width] > 0
          t += options[:cols][name][:width]
          max_width[name] = options[:cols][name][:width] - options[:gap]
          set_width += options[:cols][name][:width]
        else
          t += w + options[:gap]
          adjustment_width += w
          set_width += options[:gap]
        end
      end
      pos[:__end__] = t
      [pos, t, x, adjustment_width, set_width]
    end

    def __open_new_object
      save_state
      tOID = open_object
      close_object
      add_object(tOID)
      reopen_object(tOID)
      tOID
    end

    private :__find_table_max_width, :__open_new_object

      # This function will intially be used to implement underlining support,
      # but could be used for a range of other purposes
    def ez_process_text(text)
      text.gsub!(/<(\/)?[Uu]>/) { "<$1c:uline>" }
    end

      # This will add a string of +text+ to the document, starting at the
      # current drawing position. It will wrap to keep within the margins,
      # including optional offsets from the left and the right. If +size+ is
      # not specified, then it will be the last one used, or the default value
      # (12). The text will go to the start of the next line when a return
      # code "\n" is found.
      #
      # Possible +options+ are:
      # :left::            number, gap to leave from the left margin
      # :right::           number, gap to leave from the right margin
      # :aleft::           number, absolute left position (overrides :left)
      # :aright::          number, absolute right position (overrides
      #                     :right)
      # :justification::   :left, :right, :center, :centre, :full
      # :leading::         number, defines the total height taken by the
      #                     line, independent of the font height.
      # :spacing::         a real number, though usually set to one of 1,
      #                     1.5, 2 (line spacing as used in word processing)
      #
      # Only one of :leading or :spacing should be specified (leading
      # overrides spacing).
      #
      # If +test+ is +true+, then this should just check to see if the text is
      # flowing onto a new page or not; returns +true+ or +false+.
      #
    def ez_text(text, size = 0, options = nil, test = false)
        # apply the filtering which will make the underlining function.
      ez_process_text(text)

      options ||= {}

      newPage = false
      store_y = @y

      left = options[:aleft]
      left ||= @ez[:leftMargin] + (options[:left].nil? ? 0 : options[:left])

      right = options[:aright]
      right ||= @ez[:pageWidth] - @ez[:rightMargin] - (options[:right].nil? ? 0 : options[:right])

      size = @ez[:fontSize] if size <= 0
      @ez[:fontSize] = size

      just = options[:justification]
      just ||= :left

      if not options[:leading].nil? # leading instead of spacing
        height = options[:leading]
      elsif not options[:spacing].nil?
        height ||= options[:spacing] * get_font_height(size)
      else
        height = get_font_height(size)
      end

      text.each do |line|
        start = true
        loop do # while not line.empty? or start
          break if (line.nil? or line.empty?) and not start
          start = false
          @y -= height
          if @y < @ez[:bottomMargin]
            if test
              newPage = true
            else
              ez_new_page # and then re-calc the left and right, in case they have changed due to columns
            end
          end

          if not options[:aleft].nil?
            left = options[:aleft]
          else
            left = @ez[:leftMargin] + (options[:left].nil? ? 0 : options[:left])
          end

          if not options[:aright].nil?
            right = options[:aright]
          else
            right = @ez[:pageWidth] - @ez[:rightMargin] - (options[:right].nil? ? 0 : options[:right])
          end

          line = add_text_wrap(left, @y, right - left, size, line, just, 0, test)
        end
      end

      @y = store_y if test
      return newPage if test
      @y
    end

    def ilink(info)
      alink(info, true)
    end

      # A callback function to support the formation of clickable links within
      # the document.
    def alink(info, internal = false)
        # the thickness of the line as a proportion of the height. also the
        # drop of the line.
      lineFactor = 0.05 

      case info[:status]
      when :start, :sol
          # The beginning of the link. This should contain the URI for the
          # link as the :p entry, and will also contain the value of
          # :nCallback.
        i = info[:nCallback]
        @ez[:links] ||= {}
        @ez[:links][i] = {
          :x         => info[:x],
          :y         => info[:y],
          :angle     => info[:angle],
          :descender => info[:descender],
          :height    => info[:height],
          :url       => info[:p]
        }

        if not internal
          save_state
          set_color(PDF::Color.new(0, 0, 1))
          set_stroke_color(PDF::Color.new(0, 0, 1))
          set_line_style(info[:height] * lineFactor)
        end
      when :end, :eol
          # The end of the link. Assume that it is the most recent opening
          # which has closed.
        i = info[:nCallback]
        start = @ez[:links][i]
          # add underlining
        if internal
          add_internal_link(start[:url], start[:x],
                            start[:y] + start[:descender], info[:x],
                            start[:y] + start[:descender] + start[:height])
        else
          a = PDF.deg2rad(start[:angle] - 90.0)
          drop = start[:height] * lineFactor * 1.5
          dropx = Math.cos(a) * drop
          dropy = -(Math.sin(a)) * drop
          line(start[:x] - drop, start[:y] - dropy, info[:x] - dropx,
                info[:y] - dropy)
          add_link(start[:url], start[:x], start[:y] + start[:descender],
                   info[:x], start[:y] + start[:descender] + start[:height])
          restore_state
        end
      end
    end

      # a callback function to support underlining
    def uline(info)
        # the thickness of the line as a proportion of the height. also the
        # drop of the line.
      lineFactor = 0.05

      case info[:status]
      when :start, :sol
          # the beginning of the underline zone
        @ez[:links] ||= {}

        i = info[:nCallback]
        @ez[:links][i] = {
          :x         => info[:x],
          :y         => info[:y],
          :angle     => info[:angle],
          :descender => info[:descender],
          :height    => info[:height],
          :url       => info[:p]
        }

        save_state
        set_line_style(info[:height] * lineFactor)
      when :end, :eol
          # The end of the link. Assume that it is the most recent opening
          # which has closed.
        i = info[:nCallback]
        start = @ez[:links][i]
          # add underlining
        a = PDF.deg2rad(start[:angle] - 90.0)
        drop = start[:height] * lineFactor * 1.5
        dropx = Math.cos(a) * drop
        dropy = -(Math.sin(a)) * drop
        line(start[:x] - drop, start[:y] - dropy, info[:x] - dropx,
             info[:y] - dropy)
        restore_state
      end
    end
  end
end
