require 'transaction/simple'
require 'pdf'
require 'thread'
require 'image_size'

require 'pdf/arc4'
require 'pdf/color'
require 'pdf/font'
require 'pdf/writer/objects'
require 'pdf/writer/action'
require 'pdf/writer/annotation'
require 'pdf/writer/catalog'
require 'pdf/writer/contents'
require 'pdf/writer/destination'
require 'pdf/writer/encryption'
require 'pdf/writer/font'
require 'pdf/writer/fontdescriptor'
require 'pdf/writer/fontencoding'
require 'pdf/writer/image'
require 'pdf/writer/info'
require 'pdf/writer/outlines'
require 'pdf/writer/page'
require 'pdf/writer/pages'
require 'pdf/writer/procset'
require 'pdf/writer/viewerpreferences'

module PDF
  class Writer
    VERSION = "0.1.1"

    DEFAULT_PAGE_SIZE = [0, 0, 612, 792]
    LINE_CAPS = [:butt, :round, :square]
    LINE_JOINS = [:miter, :round, :bevel]

      # Contains all of the PDF objects, ready for final assembly.
    attr_reader :objects

      # The number of PDF objects in the document
    def size
      @objects.size
    end

    def generate_id
      @mutex.synchronize { @current_id += 1 }
    end

    def generate_font_id
      @mutex.synchronize { @current_font_id += 1 }
    end

      # If this is called on an existing document results may be
      # unpredictable, but the existing document would be lost at minimum this
      # function is called automatically by the constructor function
    def initialize(pagesize = DEFAULT_PAGE_SIZE)
      @mutex = Mutex.new
      @current_id = 0
      @current_font_id = 0

        # Start the document
      @objects            = []
      @callbacks          = []
      @font_families      = {}
      @fonts              = {}
      @stack              = []
      @state_stack        = []
      @loose_objects      = []
      @current_text_state = ""
      @options            = {}
      @destinations       = {}
      @add_loose_objects  = {}
      @iamges             = [] 

      @catalog        = PDF::Writer::Catalog.new(self)
      @outlines       = PDF::Writer::Outlines.new(self)
      @pages          = PDF::Writer::Pages.new(self)

      @pages.mediaBox = pagesize
      @current_node   = @pages

      @procset        = PDF::Writer::Procset.new(self)
      @info           = PDF::Writer::Info.new(self)
      @page           = PDF::Writer::Page.new(self)

      @first_page     = @page

        # Also initialize the font families that are known about already
      set_font_family(:init)

      # @file_identifier = md5("xxxxxxxx#{Time.now.to_i}")
    end

    attr_accessor :catalog
    attr_accessor :pages
    attr_accessor :procset
    attr_accessor :options
    attr_accessor :destinations
    attr_accessor :info
    attr_accessor :current_page
    attr_accessor :current_node
    attr_accessor :current_contents
    attr_accessor :loose_objects
    attr_accessor :add_loose_objects

      # Allows the user to find out what the ID is of the first page that was
      # created during startup - useful if they wish to add something to it
      # later.
    attr_reader :first_page

      # Define font families, this is used to initialize the font families for
      # the default fonts and for the user to add new ones for their fonts.
      # The default bahavious can be overridden should that be desired.
    def set_font_family(family, options = nil)
      if options.nil?
        if family == :init
            # Set the known family groups. These font families will be used to
            # enable bold and italic markers to be included within text
            # streams. HTML forms will be used... <b></b> <i></i>
          @font_families["Helvetica"] = {
            'b'   => 'Helvetica-Bold',
            'i'   => 'Helvetica-Oblique',
            'bi'  => 'Helvetica-BoldOblique',
            'ib'  => 'Helvetica-BoldOblique'
          }
          @font_families['Courier'] = {
            'b'   => 'Courier-Bold',
            'i'   => 'Courier-Oblique',
            'bi'  => 'Courier-BoldOblique',
            'ib'  => 'Courier-BoldOblique'
          }
          @font_families['Times-Roman'] = {
            'b'   => 'Times-Bold',
            'i'   => 'Times-Italic',
            'bi'  => 'Times-BoldItalic',
            'ib'  => 'Times-BoldItalic'
          }
        end
      else
          # The user is trying to set a font family. Note that this can also
          # be used to set the base ones to something else.
        @font_families[family] = options unless family.nil?
      end
    end

      # set the viewer preferences of the document, it is up to the browser to
      # obey these.
    def set_preferences(label, value = 0)
      @catalog.viewerPreferences ||= PDF::Writer::ViewerPreferences.new(self)

        # This will only work if the label is one of the valid ones.
      if label.kind_of?(Hash)
        label.each { |k, v| @catalog.viewerPreferences.__send__("#{k.downcase}=".intern, v) }
      else
        @catalog.viewerPreferences.__send__("#{label.downcase}=".intern, value)
      end
    end

      # Add a link in the document to an external URL.
    def add_link(url, x0, y0, x1, y1)
      Annotation.new(self, :link, [x0, y0, x1, y1], url)
    end

      # Add a link in the document to an internal destination (ie. within the
      # document)
    def add_internal_link(label, x0, y0, x1, y1)
      Annotation.new(self, :ilink, [x0, y0, x1, y1], label)
    end

      # Set the encryption of the document. Can be used to turn it on and/or
      # set the passwords which it will have. Also the functions that the user
      # will have are set here, such as print, modify, add.
    def set_encryption(user_pass = nil, owner_pass = nil, pc = {})
      p = ["11000000"].pack("B8")
      options = { :print => 4, :modify => 8, :copy => 16, :add => 32 }

      pc.each do |k, v|
        if v and not options[k].nil?
          p += options[k]
        elsif not options[v].nil?
          p += options[v]
        end
      end

        # Implement encryption on the document.
      if @arc4.nil?
        @arc4 = PDF::ARC4.new
        u = user_pass
        o = owner_pass || user_pass
        @encrypt_obj = PDF::Writer::Encryption.new(self, :owner => o, :user => u, :password => p)
      end
    end

    def encrypted?
      not @encrypt_obj.nil?
    end

      # should be used for internal checks, not implemented as yet
    def check_all_here
    end

      # Return the PDF stream as a string.
    def to_s(debug = false)
      @options[:compression] = false if $DEBUG or debug
      @arc4.init(@encryption_key) unless @arc4.nil?

      check_all_here

      xref = []

      content = "%PDF-1.3\n%\n" # or "%PDF-1.3\n"
      pos = content.size

      objects.each do |o|
        cont = o.to_s
        content << cont
        xref << pos
        pos += cont.size
      end

      content << "\nxref\n0 #{xref.size + 1}\n0000000000 65535 f \n"
      xref.each { |x| content << "#{'%010d' % [x]} 00000 n \n" }
      content << "\ntrailer\n  << /Size #{xref.size + 1}\n     /Root 1 0 R\n     /Info #{@info.id} 0 R\n"
        # If encryption has been applied to this document, then add the marker
        # for this dictionary
      content << "/Encrypt #{@encrypt_obj.id} 0 R\n" unless @arc4.nil?
      content << "/ID[<#{@file_identifier}><#{@file_identifier}>]\n" unless @file_identifier.nil?
      content << "  >>\nstartxref\n#{pos}\n%%EOF\n";
      content
    end

    def open_font(font)
      @fonts[font] = PDF::Font.open(font)
      @fonts[font].font_num = @fonts.size
      @fonts[font]
    end

    def pw_font(fontname)
      name = File.basename(fontname, ".afm")
      @objects.detect { |o| o.kind_of?(PDF::Writer::Font) and %r[#{o.basefont}$] =~ name }
    end

      # If the font is not loaded then load it and make the required object,
      # otherwise just make it the current font the encoding Hash can contain
      # :encoding=> 'none', 'WinAnsiEncoding', 'MacRomanEncoding' or
      # 'MacExpertEncoding' note that :encoding => 'none' will need to be used for
      # symbolic fonts and :differences => an array of mappings between numbers
      # 0->255 and character names.
    def select_font(fontname, encoding = nil, set = true)
      if @fonts[fontname].nil?
        font = open_font(fontname) # load the file
        if @fonts[fontname]
          name = File.basename(fontname).gsub(/\.afm$/, "")
          if encoding.kind_of?(Hash)
            encd = { :encoding => encoding[:encoding],
                     :differences => encoding[:differences] }.delete_if { |k, v| v.nil? }
          elsif not encoding.nil?
            encd = { :encoding => encoding }
          else
            encd = 'WinAnsiEncoding'
          end
          wfo = PDF::Writer::Font.new(self, name, encd)

            # If this is a ".afm" font, and there is a ".pfa" file to go with
            # it (as there should be for all non-basic fonts), then load it
            # into an object and put the references into the font object.
          base = font.path.sub(/\.afm$/, "")
          if File.exists?("#{base}.pfb")
            fbtype = "pfb"
          elsif File.exists?("#{base}.ttf")
            fbtype = "ttf"
          else
            fbtype = nil
          end

          fbfile = "#{base}.#{fbtype}"
            # pfbfile = "#{base}.pfb"
            # ttffile = "#{base}.ttf"
          if fontname =~ /afm/ and not fbtype.nil?
            adobefontname = font.fontname
              # Find the array of font widths, and put that into an object.
            first_char = -1;
            last_char = 0;
            widths = {}
            font.c.each_value do |details|
              num = details["C"]

              if num >= 0
                # warn "Multiple definitions of #{num}" if widths.has_key?(num)
                widths[num] = details['WX']
                first_char = num if num < first_char or first_char < 0
                last_char = num if num > last_char
              end
            end

              # Adjust the widths for the differences array.
            if encd[:differences]
              encd[:differences].each do |cnum, cname|
                # warn "Differences is ignored for now."
                (cnum - last_char).times { widths << 0 } if cnum > last_char
                last_char = cnum
                widths[cnum - firstchar] = fonts.c[cname]['WX'] if font.c[cname]
              end
            end

            widthid = PDF::Writer::Contents.new(self, :raw)
            widthid << "["
            (first_char .. last_char).each do |ii|
              if widths.has_key?(ii)
                widthid << " #{widths[ii].to_i}"
              else
                widthid << " 0"
              end
            end
            widthid << "]"

              # Load the pfb file, and put that into an object too. Note that
              # PDF supports only binary format Type 1 font files, though
              # there is a simple utility to convert them from pfa to pfb.
            data = nil
            File.open(fbfile, "rb") { |f| data = f.read }

              # create the font descriptor
            fdsc = PDF::Writer::FontDescriptor.new(self)
              # Raw contents causes problems with Acrobat Reader
            pfbc = PDF::Writer::Contents.new(self)

              # determine flags (more than a little flakey, hopefully will not
              # matter much)
            flags = 0
            flags += 2 ** 6 if font.italicangle != 0
            flags += 1 if font.isfixedpitch == "true"
            flags += 2 ** 5 # Assume a non-symbolic font

            list = { 'Ascent' => 'Ascender', 'CapHeight' => 'CapHeight',
                     'Descent' => 'Descender', 'FontBBox' => 'FontBBox',
                     'ItalicAngle' => 'ItalicAngle'}
            fdopt = { 'Flags'     => flags,
                      'FontName'  => adobefontname,
                      'StemV'     => 100 # Don't know what the value for this should be!
                    }
            list.each do |k, v|
              z = font.__send__(v.downcase.intern)
              fdopt[k] = z unless z.nil?
            end

              # Determine the cruicial lengths within this file
            if fbtype == "pfb"
              fdopt['FontFile'] = pfbc.id
              i1 = data.index('eexec') + 6
              i2 = data.index('00000000')  - i1
              i3 = data.size - i2 - i1
              pfbc.add('Length1' => i1, 'Length2' => i2, 'Length3' => i3)
            elsif fbtype == 'ttf'
              fdopt['FontFile2'] = pfbc.id
              pfbc.add('Length1' => data.size)
            end

            fdsc.options = fdopt
              # Embed the font program
            pfbc << data

              # Tell the font object about all this new stuff
            tmp = { 'BaseFont'  => adobefontname, 'Widths' => widthid.id,
                    'FirstChar' => first_char, 'LastChar'  => last_char,
                    'FontDescriptor' => fdsc.id }
            tmp['SubType'] = 'TrueType' if fbtype == "ttf"

            tmp.each { |k, v| wfo.__send__("#{k.downcase}=".intern, v) }
          end

            # Also set the differences here. Note that this means that these
            # will take effect only the first time that a font is selected,
            # else they are ignored.
          font.differences = encd[:differences] unless encd[:differences].nil?
        end
      end

      if (set and not @fonts[fontname].nil?)
          # If for some reason the font was not set in the last one then it will
          # not be selected
        @current_base_font = fontname
          # the next line means that if a new font is selected, then the
          # current text state will be applied to it as well.
        set_current_font
      end
      @current_font
    end

      # Sets up the current font, based on the font families, and the current
      # text state. Note that this system is quite flexible, a <b><i> font can
      # be completely different to a <i><b> font, and even <b><b> will have to
      # be defined within the family to have meaning. This function is to be
      # called whenever the current_text_state is changed, it will update the
      # current_font setting to whatever the appropriatte family one is. If
      # the user calls select_font themselves then that will reset the
      # current_base_font, and the current_font. This function will change the
      # current_font to whatever it should be, but will not change the
      # current_base_font.
    def set_current_font
      select_font("pdf/fonts/Helvetica.afm") if @current_base_font.nil?

      cf = File.basename(@current_base_font)
      if (not @current_text_state.nil?) and
         (not @font_families[cf].nil?) and
         (not @font_families[cf][@current_text_state].nil?)
          # Then we are in some state or another and this font has a family,
          # and the current setting exists within it select the font, then
          # return it.
        nf = File.join(File.dirname(@current_base_font), @font_families[cf][@current_text_state])
        select_font(nf, nil, false)
        @current_font = nf
      else
        @current_font = @current_base_font
      end
    end

      # add content to the currently active object
    def add_content(c)
      @current_contents << c
    end

      # sets the colour for stroke operations
    def set_color(colour, force = false)
      if force or (colour != @current_colour)
        @current_contents << ("\n%.3f %.3f %.3f rg" % [colour.r, colour.g, colour.b])
        @current_colour = colour
      end
    end

    def set_stroke_color(colour, force = false)
      if force or (colour != @current_stroke_colour)
        @current_contents << ("\n%.3f %.3f %.3f RG" % [colour.r, colour.g, colour.b])
        @current_stroke_colour = colour
      end
    end

      # draw a line from one set of coordinates to another
    def line(x1, y1, x2, y2)
      @current_contents << ("\n%.3f %.3f m %.3f %.3f l S" % [x1, y1, x2, y2])
    end

      # draw a bezier curve based on 4 control points
    def curve(x0, y0, x1, y1, x2, y2, x3, y3)
        # In the current line style, draw a bezier curve from (x0, y0) to (x3,
        # y3) using the other two points as the control points for the curve.
      @current_contents << "\n%.3f %.3f m %.3f %.3f %.3f %.3f %.3f %.3f c S" % [x0, y0, x1, y1, x2, y2, x3, y3]
    end

      # draw a part of an ellipse
    def part_ellipse(x0, y0, astart, afinish, r1, r2 = nil, angle = 0, nseg = 8)
      ellipse(x0, y0, r1, r2, angle, nseg, astart, afinish, false)
    end

      # draw a filled ellipse
    def filled_ellipse(x0, y0, r1, r2 = nil, angle = 0, nseg = 8, astart = 0, afinish = 360)
      ellipse(x0, y0, r1, r2, angle, nseg, astart, afinish, true, true)
    end

      # Draw an ellipse. Note that the part and filled ellipse are just
      # special cases of this function.
      #
      # Draws an ellipse in the current line style, centered at [x0, y0],
      # radii [r1, r2]. If r2 is +nil+, then a circle is drawn. nSeg is not
      # allowed to be less than 2, as this will simply draw a line (and will
      # even draw a pretty crappy shape at 2, as we are approximating with
      # bezier curves.
    def ellipse(x0, y0, r1, r2 = nil, angle = 0, nseg = 8, astart = 0, afinish = 360, close = true, fill = false)
      return if r1 == 0
      r2 = r1 if r2.nil?
      nseg = 2 if nseg < 2

      astart = PDF.deg2rad(astart)
      afinish = PDF.deg2rad(afinish)
      totalangle = afinish - astart
      dt = totalangle / nseg
      dtm = dt / 3

      if angle != 0
        a = -1 * PDF.deg2rad(angle)
        t = "\n q %.3f %.3f %.3f %.3f %.3f %.3f cm"
        t = t % [Math.cos(a), -1.0 * Math.sin(a), Math.sin(a), Math.cos(a), x0, y0]
        @current_contents << t
        x0 = y0 = 0
      end

      t1 = astart
      a0 = x0 + (r1 * Math.cos(t1))
      b0 = y0 + (r2 * Math.sin(t1))
      c0 = -(r1 * Math.sin(t1))
      d0 = r2 * Math.cos(t1)

      @current_contents << "\n%.3f %.3f m " % [a0, b0]
      (1..nseg).each do |i|
          # draw this bit of the total curve
        t1 = (i * dt) + astart
        a1 = x0 + (r1 * Math.cos(t1))
        b1 = y0 + (r2 * Math.sin(t1))
        c1 = -(r1 * Math.sin(t1))
        d1 = r2 * Math.cos(t1)
        t = "\n%.3f %.3f %.3f %.3f %.3f %.3f c"
        t = t % [(a0 + (c0 * dtm)), (b0 + (d0 * dtm)), (a1 - (c1 * dtm)), (b1 - (d1 * dtm)), a1, b1]
        @current_contents << t

        a0 = a1
        b0 = b1
        c0 = c1
        d0 = d1
      end

      @current_contents << (fill ? " f" : (close ? " s" : " S"))
      @current_contents << " Q" if angle != 0
    end

      # This sets the line drawing style.
      #
      # +width+::   The thickness of the line in user units.
      # +cap+:      The type of cap to put on the line (:butt, :round, or
      #             :square; a :square cap projects a flat end past the end
      #             of the line).
      # +join+::    :miter, :round, or :bevel.
      # +dash+::    An array which sets the dash pattern, a series of length
      #             values (the lengths of the on/off dashes). [2] represents
      #             a repeating 2-on and 2-off pattern. [2, 1] represents a
      #             2-on and 1-off pattern.
      # +phase+::   A modifier on the dash pattern, used to shift the point at
      #             which the pattern starts.
      #
    def set_line_style(width = 1, cap = nil, join = nil, dash = nil, phase = 0)
      s = ""
      s << "#{width} w " if width > 0
      s << "#{LINE_CAPS.index(cap)} J " unless cap.nil?
      s << "#{LINE_JOINS.index(join)} j " unless join.nil?
      unless dash.nil?
        s << " ["
        dash.each { |len| s << " #{len}" }
        s << " ] #{phase} d"
      end
      @current_line_style = s
      @current_contents << "\n#{s}"
    end

      # Draw a polygon. +points+ is an array of points, where each point is an
      # array:
      #
      #   [[0, 0], [1, 1], [2, 2]]
      #
    def polygon(points, fill = false)
      @current_contents << ("\n%.3f %.3f m " % [points[0][0], points[0][1]])
      points[1..-1].each { |p0, p1| @current_contents << ("%.3f %.3f l " % [p0, p1]) }
      @current_contents << (fill ? " f" : " S")
    end

      # A filled rectangle. Note that it is the width and height of the
      # rectangle which are the secondary paramaters, not the coordinates of
      # the upper-right corner.
    def filled_rectangle(x1, y1, width, height)
      @current_contents << ("\n%.3f %.3f %.3f %.3f re f" % [x1, y1, width, height])
    end

      # Draw a rectangle. Note that it is the width and height of the
      # rectangle which are the secondary paramaters, not the coordinates of
      # the upper-right corner.
    def rectangle(x1, y1, width, height)
      @current_contents << ("\n%.3f %.3f %.3f %.3f re S" % [x1, y1, width, height])
    end

      # Add a new page to the document. This also makes the new page the
      # current active object.
    def new_page(insert = false, page = nil, pos = :after)
        # If there is a state saved, then go up the stack closing them; on the
        # new page, re-open them with the right setings.
      ss = @state_stack.dup
      (@state_stack.size - 1).downto(1) { |i| restore_state(true) } unless @state_stack.empty?
      @state_stack = ss

      if insert
          # the ID from the ezPDF class is the ID of the contents of the page,
          # not the page object itself. Query that object to find the parent.
        PDF::Writer::Page.new(self, { :rpage => page, :pos => pos })
      else
        PDF::Writer::Page.new(self)
      end

        # If there is a stack saved, put it onto the page.
      1.upto(@state_stack.size - 1) { |i| save_state(i) } unless @state_stack.empty?

        # If there has been a stroke or fill colour set, transfer them.
      set_color(@current_colour, true) unless @current_colour.nil?
      set_stroke_color(@current_stroke_colour, true) unless @current_stroke_colour.nil?
      @current_contents << "\n#{@current_line_style}" unless @current_line_style.nil?

        # the call to the page object set @current_contents to the present page,
        # so this can be returned as the page id
      @current_contents
    end

      # Return the height in units of the current font in the given size.
    def get_font_height(size)
      select_font("pdf/fonts/Helvetica") if @fonts.empty?
      h = @fonts[@current_font].fontbbox[3].to_f - @fonts[@current_font].fontbbox[1].to_f
      (size * h / 1000)
    end

      # Return the font descender, this will normally return a negative
      # number. If you add this number to the baseline, you get the level of
      # the bottom of the font it is in the PDF user units.
    def get_font_descender(size)
      select_font("pdf/fonts/Helvetica") if @fonts.empty?
      hi = @fonts[@current_font].fontbbox[1].to_f
      (size * hi / 1000)
    end

      # Given a start position and information about how text is to be laid
      # out, calculate where on the page the text will end.
    def prvt_get_text_position(x, y, angle, size, wa, text)
      w = get_text_width(size, text)
      w += wa * (text.count(" "))
      a = PDF.deg2rad(angle)
      [Math.cos(a) * w + x, ((-Math.sin(a)) * w + y)]
    end

      # Wrapper function for prvt_check_text_directive1
    def prvt_check_text_directive(text, i, f)
      ret = prvt_check_text_directive1(text, i, f)
      [ret[0], ret[1], ret[2]]
    end

      # Checks if the text stream contains a control directive. If so, then
      # makes some changes and returns the number of characters involved in
      # the directive. This has been re-worked to include everything neccesary
      # to find the current writing point, so that the location can be sent to
      # the callback function if required. If the directive does not require a
      # font change, then +f+ should be set to 0.
    def prvt_check_text_directive1(text, i, f, final = false, x = 0, y = 0, size = 0, angle = 0, word_space_adjust = 0)
      directive = 0
      j = i
      if text[j] == ?<
        j += 1
        case text[j]
        when ?/
          j += 1
          return [directive, text, f, x, y] if text.size <= j
          case text[j]
          when ?b, ?i
            j += 1
            if text[j] == ?>
              p = @current_text_state.rindex(text[j - 1, 1])
              @current_text_state.slice!(p) if p # there is one to remove
              directive = j - i + 1
            end
          when ?c # this might be a callback function
            j += 1
            k = text.index(">", j)
            if not k.nil? and text[j] == ?:
                # Treat this as a callback directive
              directive = k - i + 1
              f = false
                # Split the remainder on colons to get the function name and
                # the parameter.
              func, *parm = text[j + 1, k - j - 1].split(/:/)
              parm = parm.join(":")
              directive = 0 if func.nil?
              unless func.nil?
                  # only call the function if this is the final call
                if final
                    # Need to assess the text position, calculate the text
                    # width to this point. Can use getTextWidth to find the
                    # text width, I think
                  tmp = prvt_get_text_position(x, y, angle, size,
                                               word_space_adjust, text[0, i])
                  x = tmp[0]
                  y = tmp[1]
                  info = { :x => x, :y => y, :angle => angle,
                           :status => :end, :p => parm,
                           :nCallback => @callbacks.size }
                  ret = __send__(func.intern, info)
                  if ret.kind_of?(Hash)
                    ret.each do |k, v|
                      x = rv if rk == :x
                      y = rv if rk == :y
                    end
                  end
                    # Also remove from the stack. For simplicity, take from
                    # the end.
                  @callbacks.pop
                end
              end
            end
          end
        when ?b, ?i
          j += 1
          if text[j] == ?>
            @current_text_state << text[j - 1, 1]
            directive = j - i + 1
          end
        when ?c, ?C
          noclose = true if text[j] == ?C
            # this may be a callback function
          j += 1
          k = text.index(">", j)
          if k and text[j] == ?:
              # Treat this as a callback directive
            directive = k - i + 1
            f = false
              # Split the remainder on colons to get the function name and
              # the parameter.
            func, *parm = text[j + 1, k - j - 1].split(/:/)
            parm = parm.join(":")
            directive = 0 if func.nil?
            unless func.nil?
                # only call the function if this is the final call
              if final
                  # Need to assess the text position, calculate the text
                  # width to this point. Can use getTextWidth to find the
                  # text width, I think
                tmp = prvt_get_text_position(x, y, angle, size,
                                             word_space_adjust, text[0, i])
                x = tmp[0]
                y = tmp[1]
                info = { :x => x, :y => y, :angle => angle,
                         :status => :start, :p => parm, :f => func,
                         :height => get_font_height(size),
                         :descender => get_font_descender(size),
                         :nCallback => @callbacks.size + 1 }
                  # Only add to the stack if this is a small "c"; a
                  # "start-stop" pair.
                @callbacks << info unless noclose
                ret = __send__(func.intern, info)
                if ret.kind_of?(Hash)
                  ret.each do |k, v|
                    x = rv if rk == :x
                    y = rv if rk == :y
                  end
                end
              end
            end
          end
        end
      end
      [directive, text, f, x, y]
    end

      # add text to the document, at a specified location, size and angle on
      # the page
    def add_text(x, y, size, text, angle = 0, word_space_adjust = 0)
      select_font("pdf/fonts/Helvetica") if @fonts.empty?

        # If there are any open callbacks, then they should be called, to show
        # the start of the line
      unless @callbacks.empty?
        @callbacks.reverse_each do |i|
          info = { :x => x, :y => y, :angle => angle, :status => :sol,
                   :p => i[:p], :nCallback => i[:nCallback],
                   :height => i[:height], :descender => i[:descender] }
          func = i[:f]
          __send__(func.intern, info)
        end
      end

      if angle == 0
        @current_contents << ("\nBT %.3f %.3f Td" % [x, y])
      else
        a = PDF.deg2rad(angle)
        t = "\nBT %.3f %.3f %.3f %.3f %.3f %.3f Tm"
        t = t % [Math.cos(a), -1 * Math.sin(a), Math.sin(a), Math.cos(a), x, y]
        @current_contents << t
      end

      if (word_space_adjust != 0) or not ((@word_space_adjust.nil?) and (@word_space_adjust != word_space_adjust))
        @word_space_adjust = word_space_adjust
        @current_contents << (" %.3f Tw" % word_space_adjust)
      end

      i = -1
      start = 0
      loop do
        i += 1
        break if i == text.size
        f = true
        directive, text, f = prvt_check_text_directive(text, i, f)
        if directive != 0
          if i > start
            part = text[start, i - start]
            t = " /F#{pw_font(@current_font).font_id} %.1f Tf (#{PDF.escape(part)}) Tj"
            @current_contents << (t % [size])
          end

          if f
            set_current_font
          else
            @current_contents << " ET"
            f = true
            xp = x
            yp = y
            directive, text, f, xp, yp = prvt_check_text_directive1(text, i, f, true, xp, yp, size, angle, word_space_adjust)

              # Restart the text object
            if angle == 0
              @current_contents << ("\nBT %.3f %.3f Td" % [xp, yp])
            else
              a = PDF.deg2rad(angle)
              t = "\nBT %.3f %.3f %.3f %.3f %.3f %.3f Tm"
              t = t % [Math.cos(a), -1 * Math.sin(a), Math.sin(a), Math.cos(a), xp, yp]
              @current_contents << t
            end

            if (word_space_adjust != 0) or (word_space_adjust != @word_space_adjust)
              @word_space_adjust = word_space_adjust
              @current_contents << (" %.3f Tw" % [word_space_adjust])
            end
          end
          i += directive - 1
          start = i + 1
        end
      end

      if start < text.size
        part = text[start..-1]
        t = " /F#{pw_font(@current_font).font_id} %.1f Tf (#{PDF.escape(part.chomp)}) Tj"
        @current_contents << (t % size)
      end
      @current_contents << " ET"
    end

      # Calculate how wide a given text string will be on a page, at a given
      # size. This can be called externally, but is alse used by the other
      # class functions.
    def get_text_width(size, text)
        # This function should not change any of the settings, though it will
        # need to track any directives which change during calculation, so
        # copy them at the start and put them back at the end.
      t_CTS = @current_text_state.dup

      select_font("pdf/fonts/Helvetica") if @fonts.empty?
        # converts a number or a float to a string so it can get the width
      t = text.to_s
        # hmm, this is where it all starts to get tricky - use the font
        # information to calculate the width of each character, add them up
        # and convert to user units
      w = 0
      cf = @current_font

      i = -1
      loop do
        i += 1
        break if i == t.size
        f = true
        directive, text, f = prvt_check_text_directive(text, i, f)
        if directive != 0
          if f
            set_current_font
            cf = @current_font
          end
          i += directive - 1
        else
          char = t[i, 1]
          char = t[i] if @fonts[cf].c[char].nil?
          if not @fonts[cf].differences.nil?
              # This character is being replaced by another.
            name = @fonts[cf].differences[char]
            w += @fonts[cf].c[name]['WX'] unless @fonts.c[name]['WX'].nil?
          elsif not @fonts[cf].c[char].nil?
            w += @fonts[cf].c[char]['WX']
          end
        end
      end

      @current_text_state = t_CTS.dup
      set_current_font

      (w * size / 1000)
    end

      # do a part of the calculation for sorting out the justification of the
      # text
    def prvt_adjust_wrap_text(text, actual, width, x, adjust, justification)
      case justification
      when :left
        nil
      when :right
        x += width - actual
      when :center, :centre
        x += (width - actual) / 2
      when :full
        spaces = text.count(" ")
        adjust = 0
        adjust = (width - actual) / spaces if spaces > 0
        [x, adjust]
      end
      [x, adjust]
    end

      # Add text to the page, but ensure that it fits within a certain width.
      # If it does not fit then put in as much as possible, splitting at word
      # boundaries; return the remainder. +justification+ and +angle+ can also
      # be specified for the text.
      #
      # This will display the text; if it goes beyond the width +width+, it
      # will backttrack to the previous space or hyphen and return the
      # remainder of the text.
      #
      # +justification+::   :left, :right, :center, :centre, or :full
    def add_text_wrap(x, y, width, size, text, justification = :left, angle = 0, test = false)
        # Need to store the initial text state, as this will change during the
        # width calculation, but will need to be re-set before printing, so
        # that the chars work out right
      t_CTS = @current_text_state.dup

      select_font("pdf/fonts/Helvetica") if @fonts.empty?
      return "" if width <= 0

      w = brk = brkw = 0
      cf = @current_font
      tw = width / size * 1000

      i = -1
      loop do
        i += 1
        break if i == text.size
        f = true
        directive, text, f = prvt_check_text_directive(text, i, f)
        if directive != 0
          if f
            set_current_font
            cf = @current_font
          end
          i += (directive - 1)
        else
          cOrd = cOrd2 = text[i, 1]
          cOrd2 = @fonts[cf].differences[cOrd] unless @fonts[cf].differences.nil?
          cOrd2 = text[i] if @fonts[cf].c[cOrd2].nil?

          w += @fonts[cf].c[cOrd2]['WX'] unless @fonts[cf].c[cOrd2].nil?
          if w > tw # We need to truncate this line
            if brk > 0 # There is somewhere to split
              if text[brk, 1] == " "
                tmp = text[0, brk]
              else
                tmp = text[0, brk + 1]
              end
              x, adjust = prvt_adjust_wrap_text(tmp, brkw, width, x, 0, justification)

                # Reset the text state
              @current_text_state = t_CTS.dup
              set_current_font
              add_text(x, y, size, tmp, angle, adjust) unless test
              return text[brk + 1..-1]
            else # just split before the current character
              tmp = text[0, i]
              ctmp = text[i]
              ctmp = @fonts[cf].differences[ctmp] unless @fonts[cf].differences.nil?
              tmpw = (w - @fonts[cf].c[ctmp]['WX']) * size / 1000
              x, adjust = prvt_adjust_wrap_text(tmp, brkw, width, x, 0, justification)

                # Reset the text state
              @current_text_state = t_CTS.dup
              set_current_font
              add_text(x, y, size, tmp, angle, adjust) unless test
              return text[i..-1]
            end
          end

          if text[i] == ?-
            brk = i
            brkw = w * size / 1000
          end

          if text[i, 1] == " "
            brk = i
            ctmp = text[i]
            ctmp = @fonts[cf].differences[ctmp] unless @fonts[cf].differences.nil?
            z = @fonts[cf].c[tmp].nil? ? 0 : @fonts[cf].c[tmp]['WX']
            brkw = (w - z) * size / 1000
          end
        end
      end

        # There was no need to break this line.
      justification = :left if justification == :full
      tmpw = (w * size) / 1000
      x, adjust = prvt_adjust_wrap_text(text, tmpw, width, x, 0, justification)
        # reset the text state
      @current_text_state = t_CTS.dup
      set_current_font
      add_text(x, y, size, text, angle, adjust) unless test
      return ""
    end

      # This will be called at a new page to return the state to what it was
      # on the end of the previous page, before the stack was closed down.
      # This is to get around not being able to have open 'q' across pages.
    def save_state(page_end = false)
      if page_end
          # This will be called at a new page to return the state to what it was
          # on the end of the previous page, before the stack was closed down.
          # This is to get around not being able to have open 'q' across pages
        opt = @state_stack[page_end]
        set_color(opt[:col], true) unless opt[:col].nil?
        set_stroke_color(opt[:str], true) unless opt[:str].nil?
        @current_contents << "\n#{opt[:lin]}" unless opt[:lin].nil?
        # @current_line_style = opt[:lin]
      else
        @state_stack << { :col => @current_colour,
                          :str => @current_stroke_colour,
                          :lin => @current_line_style }
      end
      @current_contents << "\nq"
    end

      # restore a previously saved state
    def restore_state(page_end = false)
      unless page_end
        s = @state_stack.pop
        unless s.nil?
          @current_colour = s[:col]
          @current_stroke_colour = s[:str]
          @current_contents << "\n#{s[:lin]}"
          @current_line_style = s[:lin]
        end
      end
      @current_contents << "\nQ"
    end

      # Make a loose object. The output will go into this object, until it is
      # closed, then will revert to the current one. This object will not
      # appear until it is included within a page. The function will return
      # the object reference.
    def open_object
      @stack << { :c => @current_contents, :p => @current_page }
      @current_contents = PDF::Writer::Contents.new(self)
      @loose_objects << @current_contents
      @current_contents
    end

      # open an existing object for editing
    def reopen_object(id)
      @stack << { :c => @current_contents, :p => @current_page }
      @current_contents = id
        # if this object is the primary contents for a page, then set the
        # current page to its parent
      @current_page = @current_contents.on_page unless @current_contents.on_page.nil?
      @current_contents
    end

      # close an object
    def close_object
      unless @stack.empty?
        s = @stack.pop
        @current_contents = s[:c]
        @current_page = s[:p]
      end
    end

      # stop an object from appearing on pages from this point on
    def stop_object(id)
      o = @loose_objects.detect { |i| i.id == id.id }
      @add_loose_objects[o] = nil
    end

      # After an object has been created, it will only show if it has been
      # added, using this function.
    def add_object(id, options = :add)
      o = @loose_objects.detect { |i| i == id }

      if o and @current_contents != o
        case options
        when :all, :add
          @add_loose_objects[o] = options if options == :all
          @current_contents.on_page.contents << o unless @current_contents.on_page.nil?
        when :even
          @add_loose_objects[o] = options
          page = @current_contents.on_page
          add_object(id) if page.info.page_number % 2 == 0
        when :odd
          @add_loose_objects[o] = options
          page = @current_contents.on_page
          add_object(id) if page.info.page_number % 2 == 1
        when :next
          @add_loose_objects[o] = :all
        when :nexteven
          @add_loose_objects[o] = :even
        when :nextodd
          @add_loose_objects[o] = :odd
        end
      end
    end

      # add content to the documents info object
    def add_info(label, value = 0)
        # This will only work if the label is one of the valid ones. Modify
        # this so that arrays can be passed as well. If @label is an array
        # then assume that it is key => value pairs else assume that they are
        # both scalar, anything else will probably error.
      if label.kind_of?(Hash)
        label.each { |k, v| @info.__send__(k.downcase.intern, v) }
      else
        @info.__send__(label.downcase.intern, value)
      end
    end

      # extract an integer from a position in a byte stream
      # return the integer represented by +num+ bytes from +pos+ within
      # +data+.
    def prvt_get_bytes(data, pos, num)
      ret = 0
      (0...num).each do |i|
        ret *= 256
        ret += data[pos + i]
      end
      [ret, data]
    end

      # Add a PNG image into the document, from a file. This should work with
      # remote files.
    def add_png_from_file(file, x, y, w = 0, h = 0)
      File.open(file, "rb") do |f|
        data = ""
        data << f.read(1024) while not f.eof
      end

      header = ("%c" * 8) % [137, 80, 78, 71, 13, 10, 26, 10]
      raise TypeError, PDF::Message[:png_invalid_header] if data[0, 8] != header

        # set pointer
      p = 8
      l = data.size
        # cycle through the file, identifying chunks
      have_header = false
      info = {}
      idata = ""
      pdata = ""
      loop do
        break if p >= len
        chunk_len, data = prvt_get_bytes(data, p, 4)
        chunk_type = data[p + 4, 4]
        case chunk_type
        when 'IHDR' # The file information is here
          info['width']             = prvt_get_bytes(data, p + 8, 4)
          info['height']            = prvt_get_bytes(data, p + 12, 4)
          info['bitDepth']          = data[p + 16]
          info['colorType']         = data[p + 17]
          info['compressionMethod'] = data[p + 18]
          info['filterMethod']      = data[p + 19]
          info['interlaceMethod']   = data[p + 20]
          have_header = true;

          raise TypeError, PDF::Writer[:png_unsupp_compres] if info['compressionMethod'] != 0
          raise TypeError, PDF::Writer[:png_unsupp_filter] if info['filterMethod'] != 0
        when 'PLTE'
          pdata << data[p + 8, chunk_len]
        when 'IDAT'
          idata << data[p + 8, chunk_len]
        when 'tRNS' # This chunk can only occur once and it must occur after the PLTE chunk and before IDAT chunk
          transparency = {}
          if info['colorType'] == 3 # indexed colour, RBG
              # Corresponding to entries in the plte chunk:
              #   Alpha for palette index 0: 1 byte
              #   Alpha for palette index 1: 1 byte
              # ...etc...
              # There will be one entry for each palette entry, up until the
              # last non-opaque entry. Set up an array, stretching over all
              # palette entries which will be o (opaque) or 1 (transparent).
            transparency['type'] = 'indexed'
            num_palette = pdata.size / 3
            trans = 0
            chunk_len.downto(0) { |i| trans = i if data[p + 8 + i] == 0 }
            transparency['data'] = trans
          elsif @info['colorType'] == 0 # grayscale
              # corresponding to entries in the plte chunk
              #   Gray: 2 bytes, range 0 .. (2^bitdepth)-1
            transparency['type'] = 'indexed'
            transparency['data'] = data[p + 8 + 1]
          elsif @info['colorType'] == 2 # truecolor
              # corresponding to entries in the plte chunk
              #   Red: 2 bytes, range 0 .. (2^bitdepth)-1
              #   Green: 2 bytes, range 0 .. (2^bitdepth)-1
              #   Blue: 2 bytes, range 0 .. (2^bitdepth)-1
            transparency['r'] = prvt_get_bytes(data, p + 8, 2)  # r from truecolor
            transparency['g'] = prvt_get_bytes(data, p + 10, 2) # g from truecolor
            transparency['b'] = prvt_get_bytes(data, p + 12, 2) # b from truecolor
          end
        end

        p += chunk_len + 12
      end

      raise TypeError, PDF::Writer[:png_header_missing] unless have_header
      raise TypeError, PDF::Writer[:png_8bit_colour] if info['bitDepth'] > 8
      raise TypeError, PDF::Writer[:png_alpha_trans] unless [0, 2, 3].include?(info['colorType'])

      ncolor  = 1
      colour  = 'DeviceRGB'
      case info['colorType']
      when 2
        ncolor = 3
      when 0
        colour = 'DeviceGray'
      end

      w = h / info['height'] * info['width']
      h = w * info['height'] / info['width']
      
        # The image is okay. Add it to the document.
      t = Time.now
      @images << t
      im = @images.index(t)
      label = "I#{im}"

      options = { 'label' => label, 'bitsPerComponent' => info['bitDepth'],
                  'pdata' => pdata, 'type' => 'png', 'color' => colour,
                  'ncolor' => ncolor }
      options['transparency'] = transparency unless transparency.nil?
      @images[im] = PDF::Writer::Image.new(self, data, info['width'], info['height'], options)
      @current_contents << "\nq\n#{w} 0 0 #{h} #{x} #{x} cm\n/#{label} Do\nQ"
    end

      # Add a JPEG image into the document, from a file.
    def add_jpeg_from_file(img, x, y, w = 0, h = 0)
        # Attempt to add a jpeg image straight from a file, using no GD
        # commands. Note that this function is unable to operate on a remote
        # file.
      image = nil
      File.open(img) { |f| image = f.read }
      return if image.nil?

      is = ImageSize.new(image)
      image_width = is.get_width
      image_height = is.get_height

      # channels = is.get_channels
      channels ||= 3

      w = image_width if w <= 0 and h <= 0
      w = h / image_height * image_width if w == 0
      h = w * image_height / image_width if h == 0

      image = add_jpeg_image_common(image, x, y, w, h, image_width, image_height, channels)
    end

      # common code used by the two JPEG adding functions
    def add_jpeg_image_common(data, x, y, w = 0, h = 0, image_width = 0, image_height = 0, channels = 3)
      t = Time.now
      @images << t
      im = @images.index(t)
      label = "I#{im}"

      @images[im] = PDF::Writer::Image.new(self, data, image_width, image_height, 'label' => label, 'channels' => channels)
      @current_contents << "\nq\n#{w} 0 0 #{h} #{x} #{y} cm\n/#{label} Do\nQ"
    end

      # Specify where the document should open when it first starts. +style+
      # must be one of the following:
      #
      # ['XYZ']   a:left, b:top, c:zoom
      # ['Fit']
      # ['FitH']  a:top
      # ['FitV']  a:left
      # ['FitR']  a:left, b:bottom, c:right
      # ['FitB']
      # ['FitBH'] a:top
      # ['FitBV'] a:left
    def open_here(style, a = 0, b = 0, c = 0)
      @catalog.openHere = PDF::Writer::Destination.new(self, @current_page, style, a, b, c)
    end

      # create a labelled destination within the document
      #
      # ['XYZ']   a:left, b:top, c:zoom
      # ['Fit']
      # ['FitH']  a:top
      # ['FitV']  a:left
      # ['FitR']  a:left, b:bottom, c:right
      # ['FitB']
      # ['FitBH'] a:top
      # ['FitBV'] a:left
    def add_destination(label, style, a = 0, b = 0, c = 0)
      @destinations[label] = PDF::Writer::Destination.new(self, @current_page, style, a, b, c)
    end

    include Transaction::Simple
  end
end
