module PDF
  class Writer
      # object which is a parent to the pages in the document
    class Pages < PDF::Writer::WriterObject
      def initialize(parent)
        super(parent)

        @parent.catalog.pages = self

        @pages    = []
        @procset  = nil
        @mediaBox = nil
        @fonts    = []
        @xObjects = []
      end

      def size
        @pages.size
      end

      def first_page
        @pages[0]
      end

        # Add the page ID to the end of the page list.
      def <<(p)
        if p.kind_of?(PDF::Writer::Page)
          @pages << p
        elsif p.kind_of?(PDF::Writer::Font)
          @fonts << p
        elsif p.kind_of?(PDF::Writer::ExternalObject)
          @xObjects << p
        else
          raise ArgumentError, PDF::Message[:req_FPXO]
        end
      end

        # Add a page to the page list. If p is just a Page, then it will be
        # added to the page list. Otherwise, it will be treated as a Hash with
        # keys :page, :pos, and :rpage. :page is the Page to be added to the
        # list; :pos is :before or :after; :rpage is the Page to which the
        # new Page will be added relative to.
      def add(p)
        if p.kind_of?(PDF::Writer::Page)
          @pages << p
        elsif p.kind_of?(PDF::Writer::Font)
          @fonts << p
        elsif p.kind_of?(PDF::Writer::ExternalObject)
          @xObjects << p
        elsif p.kind_of?(Hash)
            # Find a match.
          i = @pages.index(p[:rpage])
          unless i.nil?
              # There is a match; insert the page.
            case p[:pos]
            when :before
              @pages[i, 0] = p[:page]
            when :after
              @pages[i + 1, 0] = p[:page]
            else
              raise ArgumentError, PDF::Message[:invalid_pos]
            end
          end
        else
          raise ArgumentError, PDF::Message[:req_FPXOH]
        end
      end

      attr_accessor :procset
      attr_accessor :mediaBox # should be an array of 4 numbers

      def to_s
        unless @pages.empty?
          res = "\n#{@id} 0 obj\n<< /Type /Pages\n/Kids ["
          @pages.uniq! # uniqify the data...
          @pages.each { |p| res << "#{p.id} 0 R\n" }
          res << "]\n/Count #{@pages.size}"
          unless @fonts.empty? and @procset.nil?
            res << "\n/Resources <<"
            res << "\n/ProcSet #{@procset.id} 0 R" unless @procset.nil?
            unless @fonts.empty?
              res << "\n/Font << "
              @fonts.each { |f| res << "\n/F#{f.font_id} #{f.id} 0 R" }
              res << " >>"
            end
            unless @xObjects.empty?
              res << "\n/XObject << "
              @xObjects.each { |x| res << "\n/#{x.label} #{x.id} 0 R" }
              res << " >>"
            end
            res << "\n>>"
            res << "\n/MediaBox [#{@mediaBox.join(' ')}]" unless @mediaBox.nil?
          end
          res << "\n >>\nendobj"
        else
          "\n#{@id} 0 obj\n<< /Type /Pages\n/Count 0\n>>\nendobj"
        end
      end
    end
  end
end
